#include <stdio.h>
#include <portaudio.h>

#include "module_utils.h"

#include "pac_queue.h"
#include "ringbuffer.h"
#include "synthesizer.h"
#include "player.h"
#include "pac_logger.h"

#ifndef min
	#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

#define SAMPLE_RATE 48000

PaStream *stream;
PaError err;

RingBuffer *rbuf = NULL;
size_t MAX_RBUF_SIZE = 4*2048;

typedef struct {
	PWave *crr_wave;
	size_t crr_index;
	gboolean empty;
} wave_data;

wave_data *pWData;

sem_t player_stop_semaphore;
sem_t buffer_finished;

pthread_t player_thread;
void *_player(void *nothing);

pthread_t stop_thread;
void *_stop(void *nothing);

int pac_pa_callback(const void *input_buffer, void *output_buffer,
		    unsigned long frames_per_buffer, 
		    const PaStreamCallbackTimeInfo *time_info,
		    PaStreamCallbackFlags status_flags, 
		    void *user_data);

gboolean player_init()
{
	
	//Here we must initialize portaudio
/*	err = Pa_Initialize();
	if(err != paNoError)
	{
		PAC_LOG("player : portaudio eror : %s", Pa_GetErrorText(err));
	}
	
	pWData = g_new(wave_data, 1);
	pWData->crr_wave = NULL;
	pWData->crr_index = 0;
	pWData->empty = TRUE;
	
	err = Pa_OpenDefaultStream( &stream,
		 		    0,
				    1,
				    paInt16,
				    SAMPLE_RATE,
				    paFramesPerBufferUnspecified,
				    pac_pa_callback,
				    pWData); 		*/
	//Pa_StartStream(stream);	
	PAC_LOG("player : portaudio init done!");
//	rbuf = g_new(RingBuffer, 1);
//	void *rbuf_data = malloc(MAX_RBUF_SIZE);
//	RingBuffer_Init(rbuf, MAX_RBUF_SIZE, rbuf_data);	
	PAC_LOG("RingBuffer init done!");	
	
	int ret = pthread_create(&player_thread, NULL, _player, NULL);
	if(ret != 0)
	{
		PAC_LOG("Can't create player thread!!!");
		return FALSE;
	}
	
//	Pa_StartStream(stream);
	
	ret = pthread_create(&stop_thread, NULL, _stop, NULL);
	if(ret != 0)
	{
		PAC_LOG("Can't create player_stop thread!!!");
		return FALSE;
	}
	
	sem_init(&player_stop_semaphore, 0, 0);
	sem_init(&buffer_finished, 0, 0);

	return TRUE;
}

void player_close()
{
	/*err = Pa_StopStream(stream);
	if(err != paNoError)
	{
		PAC_LOG("player : portaudio error : %s", Pa_GetErrorText(err));
	}

	err = Pa_Terminate();
	if(err != paNoError)
	{
		PAC_LOG("player : portaudio error : %s", Pa_GetErrorText(err));
	}
*/
	pthread_cancel(player_thread);
	pthread_join(player_thread, NULL);
	
	pthread_cancel(stop_thread);
	pthread_join(stop_thread, NULL);
	
	sem_destroy(&player_stop_semaphore);	

	RingBuffer_Flush(rbuf);
	free(rbuf->buffer);
	g_free(rbuf);
}

void *_player2(void *nothing)
{

	long available = 0;
	PWave *crr_audio = NULL;
	size_t crr_audio_index = 0;

	set_speaking_thread_parameters();
	
	while(1)
	{
		PAC_LOG("_player : first while ,,,, ");	
		pWData->crr_wave = pactos_queue_pop(sample_queue);
		pWData->crr_index = 0;
		pWData->empty = FALSE;
		
		if(Pa_IsStreamStopped(stream))
		{
			Pa_StartStream(stream);
		}

		module_report_event_begin();	

		//sem_wait(&buffer_finished);
		while(pWData->empty == FALSE) //Access to shared variable ...
					      //... without synchronization!
		{
			sleep(0.0001);//PAC_LOG("heeeeeeeeeeyyyyyyyyyyeee!!!");
		}
		
		module_report_event_end();
		if(pWData->crr_wave)
		{
			g_free(pWData->crr_wave->samples);
			g_free(pWData->crr_wave);
			pWData->crr_wave = NULL;
		}
		
	}
}

void *_player(void *nothing)
{
	AudioTrack track;

	set_speaking_thread_parameters();
	PWave *crr_wave = NULL;	
	while(1)
	{
		PAC_LOG("_player : first while ,,,, ");	
		if(crr_wave)
		{
			g_free(crr_wave->samples);
		}
		PWave *crr_wave = pactos_queue_pop(sample_queue);
		
		track.num_samples = crr_wave->num_samples;
		track.num_channels = 1;
		track.sample_rate = crr_wave->sampling_freq;
		track.bits = 16;
		track.samples = crr_wave->samples;			

		if(crr_wave->event_begin)
		{
			module_report_event_begin();	
		}
		
		int ret = module_tts_output(track, SPD_AUDIO_LE);
		if(ret < 0)
		{
			PAC_LOG("Can't play track for unknown reason.");
			return FALSE;
		}
		//sem_wait(&buffer_finished);
		
		if(crr_wave->event_end)
		{
			module_report_event_end();
		}
		/*if(crr_wave)
		{
			g_free(crr_wave->samples);
		}*/
		
	}
}

void *_stop(void *nothing)
{
	set_speaking_thread_parameters();
	while(1)
	{
		sem_wait(&player_stop_semaphore);
		
		if(module_audio_id)
		{
			PAC_LOG("stopping audio");
			int ret = spd_audio_stop(module_audio_id);
			if(ret != 0)
			{
				PAC_LOG("WARNING : Non 0 value from aydio stop!");
			}
		}	
	}
}

int pac_pa_callback(const void *input_buffer, void *output_buffer,
		    unsigned long frames_per_buffer, 
		    const PaStreamCallbackTimeInfo *time_info,
		    PaStreamCallbackFlags status_flags, 
		    void *user_data)
{
	wave_data *data = (wave_data *)user_data;
	if(data->crr_wave == NULL)
		return 0;
	signed short *out = (signed short *)output_buffer;
	size_t idx = data->crr_index;
	signed short *samples = &(data->crr_wave->samples[idx]); 

	for(size_t smp = 0; smp < frames_per_buffer; smp++)
	{
		if(data->crr_index >= data->crr_wave->num_samples)
		{
			data->empty = TRUE;
			//sem_post(&buffer_finished);
//			module_report_event_end();
			return 0;
		}
		*out++ = *samples++;
		data->crr_index++;
	}

//	RingBuffer_Read(rbuf, output_buffer, frames_per_buffer);
	
	return 0;
}






