
#include <HTS_engine.h>

#include "module_utils.h"

#include "label_generator.h"
#include "pac_logger.h"
#include "synthesizer.h"

#define LANG_EN 1
#define LANG_FA 2

char *hts_voice_file_fa = "/usr/local/share/speech-dispatcher/pactos_tavakoli.htsvoice";
char *hts_voice_file_en = "/usr/local/share/speech-dispatcher/cmu_us_arctic_slt.htsvoice";

PQueue *sample_queue = NULL;
HTS_Engine engine_fa, engine_en, *p_engine;
pthread_t generate_sample_thread;

size_t wave_counter = 0;

void *_generate_samples(void *nothing);

gboolean synthesizer_init()
{	
	HTS_Engine_initialize(&engine_fa);
	HTS_Engine_initialize(&engine_en);
	
	HTS_Engine_set_audio_buff_size(&engine_fa, 0);	
	HTS_Engine_set_audio_buff_size(&engine_en, 0);	
	
	if(!HTS_Engine_load(&engine_fa, &hts_voice_file_fa, 1))
	{
		PAC_LOG("Can't load hts voice file!!!");
		return FALSE;
	}
	if(!HTS_Engine_load(&engine_en, &hts_voice_file_en, 1))
	{
		PAC_LOG("Can't load english hts voice file!!!");
		return FALSE;
	}
	
	sample_queue = pactos_queue_create();
	
	int ret = pthread_create(&generate_sample_thread, NULL, _generate_samples, NULL);
	
	if(ret != 0)
	{
		PAC_LOG("Can't create sample generator thread!!!");
		return FALSE;
	}
	
	return TRUE;
}

void synthesizer_close()
{
	HTS_Engine_clear(&engine_fa);
	HTS_Engine_clear(&engine_en);
	
	pthread_cancel(generate_sample_thread);
	pthread_join(generate_sample_thread, NULL);

	pactos_queue_clear(sample_queue);	
}


void *_generate_samples(void *nothing)
{
	char ret = -1;
	HLabel *crr_label;
	PWave *crr_wave;

	set_speaking_thread_parameters();

	while(1)
	{
		crr_label = pactos_queue_pop(label_queue);
		if(crr_label->size < 3)
		{
			
			free_label(crr_label->label, crr_label->size);
			g_free(crr_label);
			crr_label = NULL;
			continue;
		}		
		//maybe set the hts synthesize parameters here

		p_engine = (crr_label->lang == LANG_FA) ? &engine_fa : &engine_en;
		ret = HTS_Engine_synthesize_from_strings(p_engine,
							 crr_label->label, crr_label->size);
		if(!ret)
		{
			PAC_LOG("Can't synthesize labels!!!");	
			free_label(crr_label->label, crr_label->size);
			g_free(crr_label);
			crr_label = NULL;
			continue;
		}
		size_t nsamples = HTS_Engine_get_nsamples(p_engine);
		PAC_LOG("We have generated %lu samples.", nsamples);

		signed short *generated_samples	= g_new(signed short, nsamples);
		if(generated_samples == NULL)
		{
			PAC_LOG("Can't allocate memory for generated samples!!!");
			free_label(crr_label->label, crr_label->size);
			g_free(crr_label);
			crr_label = NULL;
		}
		
		for(size_t i = 0; i < nsamples; i++)
		{
			double x = HTS_Engine_get_generated_speech(p_engine, i);
			if(x > 32767.0)
				generated_samples[i] = 32767;
			else if(x < -32768.0)
				generated_samples[i] = -32768;
			else
				generated_samples[i] = (signed short) x;
		}
		
		crr_wave = g_new(PWave, 1);
		crr_wave->samples = generated_samples;
		crr_wave->num_samples = nsamples;
		crr_wave->sampling_freq = HTS_Engine_get_sampling_frequency(p_engine);
		crr_wave->event_begin = crr_label->event_begin;
		crr_wave->event_end = crr_label->event_end;
		
		pactos_queue_push(sample_queue, crr_wave);
		
		HTS_Engine_refresh(p_engine);
		free_label(crr_label->label, crr_label->size);
		g_free(crr_label);
		crr_label = NULL;
	}
}
