#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <ctype.h>

#include "module_utils.h"

#include "text_reader.h"
#include "player.h"
#include "pac_logger.h"

extern char **pactos_text_message;
PQueue *text_queue = NULL;
PQueue *raw_text_queue = NULL;

FILE *input_file = NULL;

sem_t text_reader_semaphore;
pthread_t text_reader_thread;
pthread_t text_splitter_thread;

char text_buffer[1000] = {0};

#define LANG_EN 1
#define LANG_FA 2

void *_split_text(void *nothing);

static gboolean str_is_space(char *str)
{
	char *p_str = str;
	while(*p_str)
	{
		if(!isspace(*p_str++))
			return FALSE;
	}

	return TRUE;
}

static GArray *split_text_to_langs2(char *text, size_t len)
{
	PAC_LOG("split_text started ...");

	GArray *res = g_array_new(FALSE,FALSE,sizeof(GString *));

	GString *crr = g_string_new(NULL);

	PAC_LOG("input text is NULL? %s , len == %lu\n", text?"no":"yes", strlen(text));

	char *p_text = text;
	GUnicodeScript crr_script = g_unichar_get_script(g_utf8_get_char(p_text));

	GUnicodeScript prev_script = -1, pprev_script = -1;
	gboolean string_saved = FALSE;
	while(*p_text)
	{
		gunichar u_chr = g_utf8_get_char(p_text);

		crr_script = g_unichar_get_script(u_chr);
		if( crr_script == prev_script
			|| crr_script == G_UNICODE_SCRIPT_COMMON
			|| crr_script == G_UNICODE_SCRIPT_INHERITED
			|| (prev_script == G_UNICODE_SCRIPT_COMMON && pprev_script == -1)
			|| (prev_script == G_UNICODE_SCRIPT_INHERITED && pprev_script == -1)
			|| (prev_script == G_UNICODE_SCRIPT_COMMON && pprev_script == crr_script)
			|| (prev_script == G_UNICODE_SCRIPT_INHERITED && pprev_script == crr_script)

		  )
		{
			g_string_append_unichar(crr, u_chr);
			p_text = g_utf8_next_char(p_text);
			pprev_script = prev_script;
			prev_script = crr_script;
			string_saved = FALSE;
			continue;
		}
		g_string_append_c(crr, 0);
		g_array_append_val(res,crr);

		string_saved = TRUE;

		crr = g_string_new(NULL);
		g_string_append_unichar(crr, u_chr);

		p_text = g_utf8_next_char(p_text);

		pprev_script = prev_script;
		prev_script = crr_script;

	}

	if(!string_saved)
	{
		if(res->len == 0)
		{
			g_string_append_c(crr, 0);
			g_array_append_val(res,crr);
		}
		else
		{
			char *head_str = g_array_index(res,GString *,res->len-1)->str;
			GUnicodeScript head_script = -1;

			while(head_script < 2)
			{
				PAC_LOG("split_text : while(head_script < 2)...");
				head_script = g_unichar_get_script(g_utf8_get_char(head_str));
				head_str = g_utf8_next_char(head_str);
			}
			GUnicodeScript unsaved_str_script = -1;

			head_str = crr->str;
			while(unsaved_str_script < 2)
			{
				unsaved_str_script = g_unichar_get_script(g_utf8_get_char(head_str));

				head_str = g_utf8_next_char(head_str);
			}

			if(head_script == unsaved_str_script)
			{
				GString *p_str = g_array_index(res,GString *,res->len-1);

				g_string_erase(p_str, p_str->len-1,1);

				g_string_append(p_str,crr->str);
				g_string_append_c(p_str, 0);

			}
			else
			{
				g_string_append_c(crr, 0);
				g_array_append_val(res,crr);
			}
		}
	}

	size_t arr_len = res->len;
	for(int idx = 0; idx < arr_len ; idx++)
	{
		if(str_is_space(g_array_index(res, GString *, idx)->str))
		{
			g_array_remove_index(res,idx--);
		}
	}
	
	PAC_LOG("split_text finished.");

	return res;
}

static int  gen_scripts(char *text, int **scripts)
{
	char *p_text = text;
	size_t utf8_len = g_utf8_strlen(text, -1);
	*scripts = g_new(int, utf8_len);
	int *p_scripts = *scripts;
	size_t idx = 0;
	while(*p_text)
	{
		p_scripts[idx++] = g_unichar_get_script(g_utf8_get_char(p_text));
		p_text = g_utf8_next_char(p_text);
	}

	return utf8_len;
}

GArray *split_text_to_langs(char *text)
{
	GArray *res = g_array_new(FALSE,FALSE,sizeof(GString *));
	GString *crr = g_string_new(NULL);

	int *scripts;
	char *p_text = text;

	int n_scripts = gen_scripts(p_text, &scripts);
	size_t idx = 0;
	int nonzero = -10;
	while(idx < n_scripts)
	{
		if(scripts[idx] > 1)
		{
			nonzero = scripts[idx];
			break;
		}
		idx++;
	}
	if(nonzero < 0)
	{
		if(scripts) g_free(scripts);
		return NULL;
	}
	if(scripts[0] < 2 && scripts[0] > -1)
	{
		scripts[0] = nonzero;
	}
	for(size_t i = 1; i < n_scripts; i++)
	{
		if(scripts[i] < 2 && scripts[i] > -1)
		{
			scripts[i] = scripts[i-1];
		}
	}

	p_text = text;
	gunichar u_chr = g_utf8_get_char(p_text);
	g_string_append_unichar(crr, u_chr);
	p_text = g_utf8_next_char(p_text);
	idx = 1;
	while(*p_text)
	{
		u_chr = g_utf8_get_char(p_text);
		if(scripts[idx]==scripts[idx-1])
		{
			g_string_append_unichar(crr, u_chr);
		}
		else
		{
			g_string_append_c(crr, 0);
			if(!str_is_space(crr->str))
			{
				g_array_append_val(res,crr);
			}
			crr = g_string_new(NULL);
			g_string_append_unichar(crr, u_chr);
		}

		p_text = g_utf8_next_char(p_text);
		idx++;
	}

	if(crr->len)
	{
		g_string_append_c(crr, 0);
		if(!str_is_space(crr->str))
		{
			g_array_append_val(res,crr);
		}
	}

	if(scripts)
	{
		g_free(scripts);
	}
	return res;
}

size_t get_text_script(char *txt_ptr)
{
//	char *txt_ptr = p_str->str;
	GUnicodeScript text_script = g_unichar_get_script(g_utf8_get_char(txt_ptr));
	while(text_script < 2)
	{
		txt_ptr = g_utf8_next_char(txt_ptr);
		text_script = g_unichar_get_script(g_utf8_get_char(txt_ptr));
	}
	return (text_script < 3)?LANG_FA:LANG_EN;
}


gboolean text_reader_init()
{
//	input_file = fopen(input_text_file, "r");
//	if(input_file == NULL)
//	{
//		PAC_LOG("Can't open input file : %s", input_text_file);
//		return FALSE;
//	}
	text_queue = pactos_queue_create();
	raw_text_queue = pactos_queue_create();

	int ret = pthread_create(&text_reader_thread, NULL, _read_text, NULL);
	if(ret != 0)
	{
		PAC_LOG("Can't create text reader thread!!!");
		return FALSE;
	}
	ret = pthread_create(&text_splitter_thread, NULL, _split_text, NULL);
	if(ret != 0)
	{
		PAC_LOG("Can't create text splitter thread!!!");
		return FALSE;
	}
	sem_init(&text_reader_semaphore, 0, 0);

	return TRUE;
}

void text_reader_close()
{
//	if(input_file)
//	{
//		fclose(input_file);
//	}
	pactos_queue_clear(text_queue);
	pactos_queue_clear(raw_text_queue);
	
	pthread_cancel(text_reader_thread);
	pthread_join(text_reader_thread, NULL);
	
	pthread_cancel(text_splitter_thread);
	pthread_join(text_splitter_thread, NULL);
}

void get_random_line()
{
	int count = 0;
	char line[1000] = {0};
	fseek(input_file, 0, SEEK_SET);
	while(fgets(line,1000,input_file) != NULL)
	{
		++count;
		if((rand() / (float)RAND_MAX) <= (1.0 / count))
		{
			strcpy(text_buffer, line);
		}
	}
}

void *_read_text(void *nothing)
{
//	set_speaking_thread_parameters();	
	while(1)
	{
		PAC_LOG("we are waiting new text!");
		sem_wait(&text_reader_semaphore);
		
		pactos_queue_push(raw_text_queue, *pactos_text_message);
	}
	
	return NULL;
}

void *_split_text(void *nothing)
{
//	set_speaking_thread_parameters();	

	while(1)
	{
		char *raw_text = pactos_queue_pop(raw_text_queue);
		
		GArray *splitted_text = split_text_to_langs(raw_text);
		PAC_LOG("message splitted in %lu parts.", splitted_text->len);
		for(size_t str_idx = 0; str_idx < splitted_text->len; str_idx++)
		{
			GString *crr_str = g_array_index(splitted_text, GString *, str_idx);
			PAC_LOG("part %lu : %s", str_idx, crr_str->str); 
			pactos_queue_push(text_queue, crr_str->str);
		}
		
		g_free(raw_text);
	}
}
